# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 14:31:04 2021

@author: Andreas
"""
import copy
import numpy as np


def jacobian(fun, argnum=0, h=1e-6, mode=2):
    """Decorator for the jacobian as 2- or 3-point finite-differences
    approximation w.r.t. a given argnum and h.

    Parameters
    ----------
    fun : function
        Function for which the jacobian should be approximated.
    argnum : int
        Evaluate the jacobian w.r.t the the selected argument (default is 0).
    h : float
        A small number (default is 1e-6).
    mode : int

    Returns
    -------
    jacwrapper : function
        Function for the calculation of the jacobian of function `fun`
        w.r.t. given `argnum`.
    """

    def jacwrapper(*args, **kwargs):
        """Calculates the jacobian as 2- or 3-point finite-differences
        approximation w.r.t. a given argnum and h."""

        # pre-evaluate f0 = f(x0) if 2-point scheme is used
        f0 = fun(*args, **kwargs)

        # check if arg is an array
        if isinstance(args[argnum], np.ndarray):
            # init 2d-jacobian
            nargs = np.size(args[argnum])
            nfuns = np.size(f0)
            jac = np.zeros((nfuns, nargs))

            # loop over columns
            for j in range(nargs):

                # copy args and modify item j of 1d-args
                fwdargs = copy.deepcopy(args)
                fwdargs[argnum].ravel()[j] = fwdargs[argnum].ravel()[j] + h

                f = fun(*fwdargs, **kwargs)

                # re-define f0
                if mode == 3:
                    rvsargs = copy.deepcopy(args)
                    rvsargs[argnum].ravel()[j] = rvsargs[argnum].ravel()[j] - h
                    f0 = fun(*rvsargs, **kwargs)

                jac[:, j] = (f - f0).ravel() / h / (mode - 1)

            # reshape 2d-jacobian to desired shape
            jac = jac.reshape(*f0.shape, *args[argnum].shape)

        else:  # arg is float
            # allow item assignment (convert tuple of args to list)
            fwdargs = list(copy.deepcopy(args))
            fwdargs[argnum] = fwdargs[argnum] + h

            f = fun(*fwdargs, **kwargs)

            # re-define f0
            if mode == 3:
                rvsargs = list(copy.deepcopy(args))
                rvsargs[argnum] = rvsargs[argnum] - h
                f0 = fun(*rvsargs, **kwargs)

            # calculate jacobian
            jac = (f - f0) / h / (mode - 1)

        return jac

    return jacwrapper



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 14:31:04 2021

@author: Andreas
"""
import numpy as np


def argparser(fun):
    def inner(x, *args, **kwargs):
        if args[0] is None and kwargs is None:
            f = fun(x)
        if args[0] is not None and kwargs is None:
            f = fun(x, *args)
        if args[0] is None and kwargs is not None:
            f = fun(x, **kwargs)
        else:  # args[0] is not None and kwargs is not None:
            f = fun(x, *args, **kwargs)
        return f

    return inner


class ResultObject:
    def __init__(self, fun, x0, jac, args):
        self.success = False
        self.message = "not started"
        self.x = x0.copy()
        self.fun = argparser(fun)(self.x, *args)
        self.jac = argparser(jac)(self.x, *args)
        self.iteration = 0
        self.niterations = 0
        self.status = 0


def newtonrhapson(fun, x0, jac, args=(None,), maxiter=8, tol=1e-8):
    "A simple n-dimensional Newton-Rhapson solver."

    # init result object with initial function evaluation
    res = ResultObject(fun, x0, jac, args)

    # iteration loop
    for res.iteration in range(maxiter):

        # solve linear equation system
        res.x += np.linalg.solve(res.jac, -res.fun)

        # calculate function and jacobian at updated x
        res.fun = argparser(fun)(res.x, *args)
        res.jac = argparser(jac)(res.x, *args)

        # convergence check
        if np.linalg.norm(res.fun) < tol:
            res.success = True
            if res.iteration == 1:
                res.message = "Solution converged in {0:2d} Iteration ".format(
                    res.iteration
                )
                res.status = 1
            else:
                res.message = "Solution converged in {0:2d} Iterations".format(
                    res.iteration
                )
                res.status = 1
            break

    # check if newton process failed
    if not res.success:
        if maxiter == 1:
            res.message = "Calculated linear solution because of input parameter `maxiter=1` (not converged)."
        else:
            res.message = "Newton-R. process failed."

    # set number of iterations
    res.niterations = res.iteration + 1

    return res



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 14:31:04 2021

@author: Andreas
"""
import numpy as np

from .jacobian import jacobian
from .helpers import needle, control
from .newtonrhapson import newtonrhapson


def funxt(y, n, ymax, fun, jac, jacmode, jaceps, args):
    """Extended equilibrium equations.

    Parameters
    ----------
    y : array
        1d-array of unknowns.
    n : array
        pre-evaluated needle-vector.
    ymax : array
        1d-array with max. allowed values of unknows.
    fun : function
        1d-array of equilibrium equations
    jac : function
        jacobian of equilibrium euqations (not used)
    jacmode : int
        2 or 3-point evaulation of the jacobian (not used)
    jaceps : float
        a small number to evaulate the jacobian (not used)
    args : tuple
        optional function arguments

    Returns
    -------
    array
        extended 1d-array of equilibrium equations
        with control equation
    """

    x, l = y[:-1], y[-1]
    return np.append(fun(x, l, *args), np.dot(n, (y - ymax)))


def jacxt(y, n, ymax, fun, jac, jacmode, jaceps, args):
    "Jacobian of extended equilibrium equations."
    x, l = y[:-1], y[-1]
    if jac is None:
        dfundx = jacobian(fun, argnum=0, mode=jacmode, h=jaceps)
        dfundl = jacobian(fun, argnum=1, mode=jacmode, h=jaceps)
    else:
        dfundx, dfundl = jac
    return np.vstack(
        (np.hstack((dfundx(x, l, *args), dfundl(x, l, *args).reshape(-1, 1))), n)
    )


def newtonxt(
    fun,
    jac,
    y0,
    control0,
    dymax,
    jacmode=3,
    jaceps=1e-6,
    args=(None,),
    maxiter=20,
    tol=1e-8,
):
    """Solve equilibrium equations starting from an initial solution
    with control component and max. allowed increase of unknowns."""

    n = needle(control0, len(y0))
    ymax = y0 + np.sign(control0) * dymax
    res = newtonrhapson(
        fun=funxt,
        x0=y0,
        jac=jacxt,
        args=(n, ymax, fun, jac, jacmode, jaceps, args),
        maxiter=maxiter,
        tol=tol,
    )
    res.dys = (res.x - y0) / dymax
    res.control = control(res.dys)
    return res



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 14:31:04 2021

@author: Andreas
"""

import numpy as np


def needle(j, m):
    "Return zero-vector with needle at component `j` and length `m`."
    n = np.zeros(m)
    n[abs(j) - 1] = 1
    return n


def control(x):
    "Signed index of greatest absolute component of vector x (1-indexed!)."
    j = abs(x).argmax()
    return int((j + 1) * np.sign(x[j]))



import numpy as np


def header():
    print("| Step (Cycle) | Control Comp. | Equili. | Status        |")
    print("|--------------|---------------|---------|---------------|")


def cycle(step, cycl, j0, control, status, fnorm, niterations):
    if cycl > 1:
        stp = "    "
    else:
        stp = "{0:4d}".format(step)

    if j0 != control and status == 1:
        sts = 2
    else:
        sts = status

    message = ["Failed       ", "Success ({:2d}#)".format(niterations), "Recycle      "]
    print(
        "| {0:4s}     ({1:1d}) | {2:+4d}  => {3:+4d} | {4:.1e} | {5:s} |".format(
            stp, cycl, j0, control, fnorm, message[sts]
        )
    )


def errorcontrol():
    print("")
    print("ERROR. Control component changed in last cycle.")
    print("       Possible solution: Reduce stepwidth.")


def errorfinal():
    print("")
    print("ERROR. Numerical continuation stopped.")



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 14:31:04 2021

@author: Andreas
"""

import numpy as np

from .jacobian import jacobian
from .newtonxt import newtonxt
from . import printinfo


def solve(
    fun,
    x0,
    lpf0,
    jac=None,
    args=(None,),
    dxmax=0.05,
    dlpfmax=0.05,
    control0="lpf",
    jacmode=2,
    jaceps=1e-6,
    maxsteps=80,
    maxcycles=4,
    maxiter=20,
    tol=1e-6,
):
    "Numeric continuation algorithm."

    # init initial control component
    ncomp = 1 + len(x0)
    if control0 == "lpf":
        j0 = ncomp
    else:
        j0 = control0

    # init y=(x,l)-combined quantities
    y0 = np.append(x0, lpf0)
    dymax = np.append(np.ones_like(x0) * dxmax, dlpfmax)

    # init list of results
    Res = [newtonxt(fun, jac, y0, j0, dymax, jacmode, jaceps, args, maxiter=0, tol=tol)]

    printinfo.header()

    # Step loop.
    for step in 1 + np.arange(maxsteps):
        ## pre-identification of control component
        res = newtonxt(
            fun, jac, y0, j0, dymax, jacmode, jaceps, args, maxiter=1, tol=tol
        )

        # Increment loop.
        for cycl in 1 + np.arange(maxcycles):
            res = newtonxt(
                fun, jac, y0, j0, dymax, jacmode, jaceps, args, maxiter=maxiter, tol=tol
            )
            printinfo.cycle(
                step, cycl, j0, res.control, res.status, np.linalg.norm(res.fun), res.niterations
            )

            if res.success:
                if res.control == j0:
                    y0 = res.x
                    Res.append(res)
                    break
                else:
                    if cycl == maxcycles:
                        printinfo.errorcontrol()
                        res.success = False
                    else:
                        # next cycle
                        j0 = res.control
            else:
                break

        if not res.success:
            printinfo.errorfinal()
            break
    return Res
